
import re

class VroxLexer:
    
    def __init__(self):
        self.rules = {}

    def add_keyword(self, *keywords):
        pattern = r'\b(?:' + '|'.join(map(re.escape, keywords)) + r')\b'
        self.rules["KEYWORD"] = re.compile(pattern)  # Precompile

    def add_number(self):
        self.rules["NUMBER"] = re.compile(r'\b\d+(\.\d+)?([eE][-+]?\d+)?\b')

    def add_string(self):
        self.rules["STRING"] = re.compile(r'(["\'])(?:(?=(\\?))\2.)*?\1')

    def add_identifier(self):
        self.rules["IDENTIFIER"] = re.compile(r'[a-zA-Z_$][a-zA-Z0-9_$]*')

    def add_operator(self, *operators):
        pattern = r'(?:' + '|'.join(map(re.escape, operators)) + r')'
        self.rules["OPERATOR"] = re.compile(pattern)

    def add_comment(self):
        self.rules["COMMENT"] = re.compile(r'//.*|/\*[\s\S]*?\*/')

    def add_whitespace(self):
        self.rules["WHITESPACE"] = re.compile(r'[ \t]+')

    def add_newline(self):
        self.rules["NEWLINE"] = re.compile(r'\n')

    def add_punctuation(self, *punctuations):
        pattern = r'[' + ''.join(map(re.escape, punctuations)) + r']'
        self.rules["PUNCTUATION"] = re.compile(pattern)
    
       
    
    def tokenize(self, textPath, outputPath,callback):
        # Open input and output files
        with open(textPath, 'r', encoding='utf-8') as input_file, open(outputPath, 'w', encoding='utf-8') as output_file:
            i = 0
            length = 0
            
            while True:
                # Read the next chunk from the input file
                chunk = input_file.read(1024)  # Read in chunks of 1024 bytes or adjust as needed
                if not chunk:
                    break  # End of file reached
                
                length = len(chunk)
                i = 0

                # Process the chunk
                while i < length:
                    # Skip whitespace
                    if chunk[i].isspace():
                        i += 1
                        continue
                    
                    # Try matching each rule from the current position
                    matched = False
                    for rule_name, rule_info in self.rules.items():
                        match = rule_info.match(chunk, i)
                        if match:
                            matched = True
                            # Write the token directly to the output file
                            output_file.write(f"{rule_name}: {match.group(0)}\n")
                            i += len(match.group(0))
                            break
                    
                    # If no match found, increment by 1 (to handle invalid characters)
                    if not matched:
                        i += 1
                        
            callback(outputPath)
        


